{-# LANGUAGE OverloadedStrings #-}
import Haste
import Haste.Deck
import Haste.DOM
import Styles

main = present_ def deck

deck = concat
  [ satPart
  , specPart
  , overheadPart
  ]

satPart =
  [ titled "Speculative Parallelism in Erlang" $
      "...with SAT solvers!"
  , titled "What is SAT?" $ column
      [ "\"In computer science, the Boolean Satisfiability Problem (sometimes called Propositional Satisfiability Problem and abbreviated as SATISFIABILITY or SAT) is the problem of determining if there exists an interpretation that satisfies a given Boolean formula.\""
      , sized 0.2 "- Wikipedia"
      ]
  , titled "Applications of SAT solving" $ list Unnumbered
      [ "Model checking"
      , "Planning and scheduling"
      , "Bioinformatics"
      , "...and a whole lot more"
      ]
  , titled "How does it work?" $ list Unnumbered
      [ sublist Unnumbered "**Input**: a boolean CNF formula"
          ["(a ∨ ¬b ∨ ¬c) ∧ (¬c ∨ d ∨ ¬e) ∧ (¬a ∨ ¬d ∨ b)"]
      , sublist Unnumbered "**Output**: a list of assignments that satisfy the formula"
          ["a ∧ ¬d ∧ ¬c"]
      , "...or `UNSAT`"
      ]
  , titled "An Erlang implementation" $ list Unnumbered
      [ "A *problem* is a list of *clauses*."
      , "A *clause* is a list of *literals*."
      , sublist Unnumbered "A *literal* is a *non-zero integer*" $
          [ "The integer `n` denotes *x<sub>n</sub>*"
          , "The integer `-n` denotes *¬x<sub>n</sub>*"
          ]
      ]
  ]

specPart =
  [ loneText "Different kinds of parallelism"
  , titled "Data parallelism" $ row
      [ img "sweatshop.jpg"
      , column
          [ list Unnumbered
              [ "*Same* operation over a *lot* of data"
              , "`parMap`, MapReduce, GPUs, SIMD CPU instructions, etc."
              , "Easy to exploit, but uncommon"
              ]
          ]
      ]
  , titled "Task (pipeline) parallelism" $ row
      [ img "car-factory.jpg"
      , column
          [ list Unnumbered
              [ "*Different*, independent operations over *streams of* data"
              , "Pipelined CPUs, UNIX pipes, signal processing, etc."
              , "Harder to exploit, but relatively common"
              ]
          ]
      ]
  , titled "Speculative parallelism" $ row
      [ img "search.jpg"
      , column
          [ list Unnumbered
              [ "Same operation until an answer is found"
              , "Any search problem, really"
              , "Especially sensitive to resource constraints"
              ]
          ]
      ]
  ]

overheadPart =
  [ titled "Too much parallelism is bad!" $ list Unnumbered
      [ "Each thread consumes memory"
      , "Each task switch takes time"
      , "Speculative parallelism is extra vulnerable!"
      ]
  , titled "Worker pools" $ row
      [ sized 0.1 ""
      , withAttrs [style "height" =: "90%"] $ image "pool-worker.jpg"
      , column
          [ list Unnumbered
              [ "One *worker thread* per core"
              , "Assign task to worker if available..."
              , "...otherwise do it yourself!"
              ]
          ]
      , sized 0.1 ""
      ]
  ]
